# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rich.console import Console


def is_jupyter_environment() -> bool:
    """Checks if the code is running in a Jupyter environment."""
    try:
        from IPython import get_ipython  # type: ignore

        if "IPKernelApp" in get_ipython().config:
            return True
    except (ImportError, AttributeError):
        return False
    except Exception:
        return False
    return False


def should_show_progress(show_progress: bool | None, console: Console | None = None) -> bool:
    """
    Determines if a progress bar should be displayed.

    Priority:
    1. Explicit 'show_progress' argument (True/False).
    2. Active Jupyter environment.
    3. Explicit 'console' object passed.
    4. Interactive terminal.

    Args:
        show_progress: Explicit override.
        console: Optional Rich Console instance.
    """
    if show_progress is not None:
        return show_progress

    if is_jupyter_environment():
        return True

    if console is not None:
        return True

    try:
        return sys.stdout.isatty()
    except Exception:
        return False



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer

from dorsal.cli.cache_app.build_cache_cmd import build_cache
from dorsal.cli.cache_app.clear_cache_cmd import clear_cache
from dorsal.cli.cache_app.path_cache_cmd import get_cache_path
from dorsal.cli.cache_app.optimize_cache_cmd import optimize_cache
from dorsal.cli.cache_app.prune_cache_cmd import prune_cache
from dorsal.cli.cache_app.show_cache_cmd import show_cache
from dorsal.cli.cache_app.export_cmd import export_cache_cmd


app = typer.Typer(name="cache", help="Manage local cache settings.", no_args_is_help=True)

app.command(name="build")(build_cache)
app.command(name="clear")(clear_cache)
app.command(name="path")(get_cache_path)
app.command(name="optimize")(optimize_cache)
app.command(name="prune")(prune_cache)
app.command(name="show")(show_cache)
app.command(name="export")(export_cache_cmd)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Annotated, Optional

import typer
from rich.panel import Panel
import json

logger = logging.getLogger(__name__)


def update_collection(
    ctx: typer.Context,
    collection_id: Annotated[str, typer.Argument(help="The unique ID of the collection to update.")],
    name: Annotated[
        Optional[str],
        typer.Option("--name", "-n", help="A new name for the collection."),
    ] = None,
    description: Annotated[
        Optional[str],
        typer.Option("--description", "-d", help="A new description for the collection."),
    ] = None,
    json_output: Annotated[
        bool,
        typer.Option("--json", help="Output the updated collection as a raw JSON object."),
    ] = False,
):
    """
    Update the name or description of a remote collection.
    """
    from dorsal.api.collection import update_collection as api_update_collection
    from dorsal.common.cli import get_rich_console, exit_cli, EXIT_CODE_ERROR
    from dorsal.common.exceptions import AuthError, DorsalClientError, DorsalOfflineError

    console = get_rich_console()
    palette: dict[str, str] = ctx.obj["palette"]

    try:
        # Show a status spinner only for the interactive mode
        status_message = f"Updating collection '[bold]{collection_id}[/]...'" if not json_output else ""
        with console.status(status_message):
            updated_collection = api_update_collection(
                collection_id=collection_id,
                name=name,
                description=description,
                mode="pydantic",
            )

    except typer.Exit:
        raise
    except ValueError as e:
        exit_cli(code=EXIT_CODE_ERROR, message=str(e))
    except DorsalOfflineError:
        raise
    except AuthError:
        raise
    except DorsalClientError as e:
        exit_cli(code=EXIT_CODE_ERROR, message=f"API Error: {e.message}")
    except Exception as e:
        logger.exception("An unexpected error occurred in 'collection update'")
        exit_cli(code=EXIT_CODE_ERROR, message=f"An unexpected error occurred: {e}")

    if json_output:
        console.print(updated_collection.model_dump_json(indent=2, by_alias=True, exclude_none=True))
    else:
        success_panel = Panel(
            f"✅ Collection '[bold]{updated_collection.name}[/]' updated successfully.",
            expand=False,
            title=f"[{palette.get('panel_title_success', 'bold green')}]Update Complete[/]",
            border_style=palette.get("panel_border_success", "green"),
        )
        console.print(success_panel)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from unittest.mock import MagicMock, patch
from dorsal.testing import run_model
from dorsal.common.model import AnnotationModel
from dorsal.file.model_runner import ModelRunner


class DummyModel(AnnotationModel):
    id = "dummy/model"
    version = "1.0"

    def main(self):
        pass


@pytest.fixture
def mock_runner():
    with patch("dorsal.file.model_runner.ModelRunner") as MockRunnerCls:
        runner_instance = MockRunnerCls.return_value

        base_result = MagicMock()
        base_result.error = None
        base_result.record = {"hash": "123", "media_type": "text/plain", "size": 100, "name": "f.txt"}

        pipeline_result = MagicMock()
        pipeline_result.error = None

        runner_instance.run_single_model.side_effect = [base_result, pipeline_result]

        yield runner_instance


def test_test_model_success(mock_runner):
    run_model(DummyModel, "/tmp/fake.txt")

    assert mock_runner.run_single_model.call_count == 2
    call_args = mock_runner.run_single_model.call_args_list[0]

    assert call_args.kwargs["options"]["calculate_similarity_hash"] is False


def test_test_model_base_failure(mock_runner):
    failure = MagicMock()
    failure.error = "Base failed"
    mock_runner.run_single_model.side_effect = [failure]

    result = run_model(DummyModel, "/tmp/fake.txt")
    assert result.error == "Base failed"
    assert mock_runner.run_single_model.call_count == 1


def test_test_model_dependency_check():
    from dorsal.file.dependencies import make_file_extension_dependency

    deps = make_file_extension_dependency(extensions=[".pdf"])

    base_result = MagicMock()
    base_result.error = None
    base_result.record = {"name": "f.txt", "extension": ".txt", "media_type": "text/plain"}

    with patch.object(ModelRunner, "run_single_model", side_effect=[base_result]) as mock_run:
        result = run_model(DummyModel, "/f.txt", dependencies=deps)
        assert "Dependency not met" in str(result.error)
        assert mock_run.call_count == 1


def test_ambiguous_config_error(mock_runner):
    with pytest.raises(ValueError, match="Ambiguous configuration"):
        run_model(DummyModel, "/f.txt", schema_id="open/generic", validation_model=MagicMock())


def test_open_schema_resolution(mock_runner):
    with patch("dorsal.file.validators.open_schema.get_open_schema_validator") as mock_get_val:
        mock_get_val.return_value = MagicMock()

        run_model(DummyModel, "/f.txt", schema_id="open/generic")

        call_args = mock_runner.run_single_model.call_args_list[1]
        assert call_args.kwargs["validation_model"] == mock_get_val.return_value



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Self
from pydantic import BaseModel, Field, model_validator

logger = logging.getLogger(__name__)


class _LinterEntityItem(BaseModel):
    """Internal model to parse the 'label' field from an entity object."""

    label: str


class OpenEntityExtractionLinter(BaseModel):
    """
    A linter for valid 'open/entity-extraction' records.

    Enforces data quality rules:
    - If `vocabulary` is present, every entity's `label` must be in that vocabulary.
    """

    entities: List[_LinterEntityItem] = Field(default_factory=list, description="The array of entity objects to check.")
    vocabulary: list[str] | None = Field(default=None, description="The vocabulary to check entity labels against.")

    @model_validator(mode="after")
    def check_labels_are_in_vocabulary(self) -> Self:
        """
        Enforces that all entity labels exist in the vocabulary, if one is provided.
        """
        if self.vocabulary is not None:
            vocab_set = set(self.vocabulary)
            missing_labels = sorted(list({item.label for item in self.entities if item.label not in vocab_set}))

            if missing_labels:
                display_limit = 5
                missing_str = ", ".join(f"'{label}'" for label in missing_labels[:display_limit])
                if len(missing_labels) > display_limit:
                    missing_str += f", ... (+{len(missing_labels) - display_limit} more)"

                raise ValueError(
                    f"Integrity Error: {len(missing_labels)} labels found in 'entities' are not declared in the 'vocabulary': {missing_str}."
                )

        return self



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from dorsal.file.utils.size import get_filesize, human_filesize, parse_filesize


def test_get_filesize_success(tmp_path):
    f = tmp_path / "test.txt"
    f.write_text("12345")
    assert get_filesize(str(f)) == 5


def test_get_filesize_error(tmp_path):
    with pytest.raises(OSError):
        get_filesize(str(tmp_path / "ghost.txt"))


def test_human_filesize():
    assert human_filesize(100) == "100 B"
    assert human_filesize(1024) == "1 KiB"
    assert human_filesize(1024 * 1024) == "1 MiB"
    assert human_filesize(1024 * 1024 * 1024) == "1.00 GiB"

    # Config has 'KiB': 0 decimal places, so 1.46 KiB rounds to '1 KiB'
    assert human_filesize(1500) == "1 KiB"


def test_parse_filesize_integers():
    assert parse_filesize("100") == 100
    assert parse_filesize("  500  ") == 500


def test_parse_filesize_units_si():
    assert parse_filesize("1 KB") == 1000
    assert parse_filesize("2.5 MB") == 2_500_000
    assert parse_filesize("1 GB") == 1_000_000_000


def test_parse_filesize_units_iec():
    assert parse_filesize("1 KiB") == 1024
    assert parse_filesize("1 MiB") == 1024 * 1024


def test_parse_filesize_case_insensitivity():
    assert parse_filesize("1 kib") == 1024
    assert parse_filesize("1 KIB") == 1024


def test_parse_filesize_errors():
    with pytest.raises(ValueError):
        parse_filesize("invalid")

    with pytest.raises(ValueError):
        parse_filesize("100 ZB")

    with pytest.raises(ValueError):
        parse_filesize("ten MB")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from dorsal.common.exceptions import (
    QuickHashConfigurationError,
    QuickHashFileInstabilityError,
    QuickHashFileSizeError,
)
from dorsal.file.utils import FILE_HASHER, QUICK_HASHER
from dorsal.file.cache.dorsal_cache import DorsalCache
from dorsal.file.utils.hashes import HashFunctionId
from dorsal.session import get_shared_cache

logger = logging.getLogger(__name__)


class HashReader:
    """Retrieves one or more hashes for a given file."""

    @property
    def _cache(self) -> DorsalCache:
        """Dynamically fetches the current valid shared cache."""
        return get_shared_cache()

    def get(
        self,
        file_path: str,
        hashes: list[HashFunctionId],
        skip_cache: bool = False,
    ) -> dict[str, str | None]:
        """
        Retrieves one or more hashes for a given file.

        It first attempts to fetch the hash from the shared cache. If a hash
        is not found in the cache or if caching is skipped, it will be
        calculated on-demand. Hashes are written back to cache.

        Args:
            file_path: The absolute path to the file.
            hashes: A list of hash algorithm identifiers to retrieve.
            skip_cache: If True, the cache check is bypassed and all
                        hashes are calculated directly. Defaults to False.

        Returns:
            A dictionary mapping the hash algorithm identifier to its
            hexadecimal string value. If a hash cannot be generated (e.g., TLSH
            on a small file), its value will be None.
        """
        results: dict[str, str | None] = {}
        hashes_to_process: set[HashFunctionId] = set(hashes)

        if not skip_cache:
            found_in_cache = set()
            for hash_func in hashes_to_process:
                cached_hash = self._cache.get_hash(path=file_path, hash_function=hash_func)
                if cached_hash is not None:
                    results[hash_func] = cached_hash
                    found_in_cache.add(hash_func)
            hashes_to_process -= found_in_cache

        if not hashes_to_process:
            return results

        try:
            modified_time = os.path.getmtime(file_path)
            file_size = os.path.getsize(file_path)
        except OSError as e:
            logger.error(f"Could not access file details for {file_path}: {e}")
            for hash_func in hashes_to_process:
                results[hash_func] = None
            return results

        file_hasher_algos = hashes_to_process & {"SHA-256", "BLAKE3", "TLSH"}

        if file_hasher_algos:
            try:
                calculated_hashes = FILE_HASHER.hash(
                    file_path=file_path,
                    file_size=file_size,
                    calculate_sha256="SHA-256" in file_hasher_algos,
                    calculate_blake3="BLAKE3" in file_hasher_algos,
                    calculate_tlsh="TLSH" in file_hasher_algos,
                )
                for hash_function, hash_val in calculated_hashes.items():
                    results[hash_function] = hash_val
                    if not skip_cache:
                        self._cache.upsert_hash(
                            path=file_path,
                            modified_time=modified_time,
                            hash_function=hash_function,
                            hash_value=hash_val,
                        )
                    hashes_to_process.remove(hash_function)

            except OSError as e:
                logger.error(f"Error during hashing file {file_path}: {e}")
                for hash_function in file_hasher_algos:
                    results[hash_function] = None
                    hashes_to_process.remove(hash_function)

        if "QUICK" in hashes_to_process:
            quick_hash = None
            try:
                quick_hash = QUICK_HASHER.hash(file_path=file_path, file_size=file_size)
            except (
                QuickHashFileInstabilityError,
                QuickHashFileSizeError,
                QuickHashConfigurationError,
                OSError,
            ) as e:
                logger.warning(f"Could not generate QUICK hash for {file_path}: {e}")

            results["QUICK"] = quick_hash
            if not skip_cache and quick_hash is not None:
                self._cache.upsert_hash(
                    path=file_path,
                    modified_time=modified_time,
                    hash_function="QUICK",
                    hash_value=quick_hash,
                )
            hashes_to_process.remove("QUICK")

        for hash_func in hashes_to_process:
            results[hash_func] = None

        return results


HASH_READER = HashReader()



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Self
from pydantic import BaseModel, Field, model_validator, ConfigDict, ValidationError

logger = logging.getLogger(__name__)


class _LinterLabelItem(BaseModel):
    """Internal model to parse the 'label' field from a label object."""

    label: str


class OpenClassificationLinter(BaseModel):
    """
    A linter for valid 'open/classification' records.

    Enforces data quality rules:

    - If any value in `labels` is not in `vocabulary`, this model complains.

    Example:
        ```python
        valid_data = {
            "labels": [{"label": "pear"}],
            "vocabulary": ["apple", "pear"]
        }

        invalid_data = {
            "labels": [{"label": "cheese"}],
            "vocabulary": ["apple", "pear"]
        }

        # This will execute just fine
        OpenClassificationLinter(**valid_data)

        # This will raise `pydantic.ValidationError`
        OpenClassificationLinter.model_validate(**invalid_data)
        ```
    """

    labels: List[_LinterLabelItem] = Field(default_factory=list, description="The array of label objects to check.")
    vocabulary: list[str] | None = Field(default=None, description="The vocabulary to check labels against.")

    @model_validator(mode="after")
    def check_labels_are_in_vocabulary(self) -> Self:
        """
        Enforces that all labels exist in the vocabulary, if one is provided.
        """
        if self.vocabulary is not None:
            vocab_set = set(self.vocabulary)
            missing_labels = [item.label for item in self.labels if item.label not in vocab_set]

            if missing_labels:
                missing_str = ", ".join(f"'{label}'" for label in missing_labels)
                raise ValueError(f"The following labels are not in the provided vocabulary: {missing_str}.")

        return self



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import typer
from unittest.mock import MagicMock
from typer.testing import CliRunner

from dorsal.cli.adapter_app.list_cmd import list_adapters
from dorsal.cli.themes.palettes import DEFAULT_PALETTE
from dorsal.common.exceptions import DorsalError

cli_app = typer.Typer()


@cli_app.callback()
def main_callback(ctx: typer.Context):
    ctx.obj = {"palette": DEFAULT_PALETTE}


cli_app.command(name="list")(list_adapters)

runner = CliRunner()


@pytest.fixture
def mock_list_deps(mocker, mock_rich_console):
    """
    Mocks backend dependencies for the `list` command.
    """
    mock_error_console = MagicMock()
    mocker.patch("dorsal.common.cli.get_rich_console", return_value=mock_rich_console)
    mocker.patch("dorsal.common.cli.get_error_console", return_value=mock_error_console)

    mock_get_formats = mocker.patch("dorsal.api.adapters.get_supported_formats")

    return {
        "get_formats": mock_get_formats,
        "error_console": mock_error_console,
        "rich_console": mock_rich_console,
    }


def test_list_basic_success(mock_list_deps):
    """Tests a standard successful listing of formats which should output a table."""
    mock_list_deps["get_formats"].return_value = [
        ("srt", "SubRip Text (.srt) - A widely used, plaintext subtitle format."),
        ("vtt", "WebVTT (.vtt) - W3C standard web subtitle format."),
    ]

    result = runner.invoke(cli_app, ["list", "open/audio-transcription"])

    assert result.exit_code == 0, result.output

    mock_list_deps["get_formats"].assert_called_once_with("open/audio-transcription")

    table_call_arg = mock_list_deps["rich_console"].print.call_args.args[0]
    assert type(table_call_arg).__name__ == "Table"
    assert "Supported Formats for 'open/audio-transcription'" in str(table_call_arg.title)


def test_list_empty_results(mock_list_deps):
    """Tests handling of a schema that exists but has no supported formats."""
    mock_list_deps["get_formats"].return_value = []

    result = runner.invoke(cli_app, ["list", "open/unknown-schema"])

    assert result.exit_code == 0

    warning_msg = str(mock_list_deps["error_console"].print.call_args.args[0])
    assert "No formats found for schema" in warning_msg
    assert "open/unknown-schema" in warning_msg


def test_list_missing_adapters_package(mock_list_deps):
    """Tests graceful failure when dorsalhub-adapters is missing, which raises a DorsalError."""
    mock_list_deps["get_formats"].side_effect = DorsalError("Please pip install dorsalhub-adapters to enable exports.")

    result = runner.invoke(cli_app, ["list", "open/audio-transcription"])

    assert result.exit_code != 0
    error_msg = str(mock_list_deps["error_console"].print.call_args.args[0])
    assert "Error:" in error_msg
    assert "Please pip install dorsalhub-adapters" in error_msg



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Any

from pydantic import ValidationError as PydanticValidationError  # noqa: F401
from jsonschema_rs import ValidationError as JsonSchemaValidationError  # noqa: F401

from dorsal.common import constants
from dorsal.common.constants import BASE_URL


class DorsalError(Exception):
    """Base class for all custom exceptions raised by the Dorsal library."""

    def __init__(
        self,
        message: str,
        request_url: str | None = None,
        original_exception: Exception | None = None,
    ):
        super().__init__(message)
        self.request_url = request_url
        self.original_exception = original_exception

    def __str__(self):
        base_message = super().__str__()
        return f"{base_message}"


class DorsalConfigError(DorsalError):
    """Config (dorsal.toml) error."""


class DorsalOfflineError(DorsalError):
    def __init__(self, message: str = "Operation blocked: Offline Mode is active."):
        super().__init__(message)


# == Validation Errors ==


class NotPopulatedError(DorsalError):
    """Indicates data is incomplete."""


class ValidationError(DorsalError):
    """Base validation error class."""


class UnsupportedHashError(ValidationError):
    """Unsupported file hash."""


class BatchSizeError(ValidationError):
    """The size of the request is too large."""


class SchemaFormatError(ValidationError):
    """Indicates that a JSON schema document is structurally invalid or unprocessable."""

    def __init__(
        self,
        message: str,
        schema_error_detail: str | None = None,
        original_exception: Exception | None = None,
    ):
        super().__init__(message)
        self.schema_error_detail = schema_error_detail
        self.original_exception = original_exception

    def __str__(self):
        msg = self.args[0]
        if self.schema_error_detail:
            msg += f"\nSchema Detail: {self.schema_error_detail}"
        return msg


class RecordValidationError(ValidationError):
    """Indicates records failed client-side validation against the dataset's schema."""

    def __init__(self, message: str, dataset_id: str, validation_summary: dict | None = None):
        super().__init__(message)
        self.dataset_id = dataset_id
        self.validation_summary = validation_summary

    def __str__(self):
        msg = super().__str__()
        if self.validation_summary:
            msg += (
                f"\nDataset ID: {self.dataset_id}. "
                f"Client-side validation summary: "
                f"Total: {self.validation_summary.get('total_records', 'N/A')}, "
                f"Valid: {self.validation_summary.get('valid_records', 'N/A')}, "
                f"Invalid: {self.validation_summary.get('invalid_records', 'N/A')}."
            )
        return msg


class TaggingError(ValidationError):
    """Error tagging a file."""

    info_url = constants.DOCS_URL_DORSAL_FILE_TAGS
    extra_info_template = f"\nDorsal File Tags Documentation: {info_url}"

    def __init__(self, message: str):
        super().__init__(message)

    def __str__(self):
        return f"{self.args[0]}\n\n{self.extra_info_template}"


class DuplicateTagError(TaggingError):
    """A duplicate tag exists."""


class InvalidTagError(TaggingError):
    """Tag is invalid."""


class AttributeConflictError(DorsalError):
    """
    Raised when an attribute or key cannot be set because it already
    exists and overwriting is not permitted by the operation.
    """

    def __init__(self, message: str):
        super().__init__(message=message)


class TemplateNotFoundError(DorsalError):
    """Raised when a report template or its essential files cannot be found."""

    pass


# == dorsal.api ==
class DatasetTypeError(DorsalError):
    """Raised when the type of a dataset prevents some action."""


# == /client ==
class DorsalClientError(DorsalError):
    """Dorsal Python client error."""

    info_url = constants.DOCS_URL_API_TROUBLESHOOTING
    extra_info = f"For help troubleshooting client errors, visit: {info_url}"

    def __init__(
        self,
        message: str,
        request_url: str | None = None,
        original_exception: Exception | None = None,
    ):
        super().__init__(message=message, request_url=request_url, original_exception=original_exception)
        self.message = message

    def __str__(self):
        message = f"{self.args[0]}"
        if self.request_url:
            message += f"\nRequest URL: {self.request_url}"

        message += f"\n\n{self.extra_info}"
        return message


class AuthError(DorsalClientError):
    """Error for client-side authentication or API key configuration issues."""

    info_url = constants.DOCS_URL_API_AUTH

    def __init__(self, message: str):
        super().__init__(message=message)

    def __str__(self):
        return f"{self.args[0]}"


class NetworkError(DorsalClientError):
    """Network error e.g. timeout or other connectivity issues."""

    info_url = constants.DOCS_URL_API_ERRORS_NETWORK
    extra_info = (
        f"Could not reach the API. Please check your network connection \nFor more information, visit: {info_url}"
    )

    def __init__(
        self,
        message: str,
        request_url: str | None = None,
        original_exception: Exception | None = None,
    ):
        super().__init__(message=message, request_url=request_url, original_exception=original_exception)

    def __str__(self):
        message = f"{self.args[0]}"
        if self.request_url:
            message += f"\nRequest URL: {self.request_url}"
        if self.original_exception:
            message += f"\nDetails: {type(self.original_exception).__name__}: {str(self.original_exception)}"

        message += f"\n\n{self.extra_info}"
        return message


class APIError(DorsalClientError):
    """HTTP Error (4xx/5xx)."""

    info_url = constants.DOCS_URL_API_ERRORS
    extra_info = (
        "The API returned an error. Check the status code and details for more information."
        f"\nFor more information on API error codes, visit: {info_url}"
    )

    def __init__(
        self,
        status_code: int,
        detail: str | dict,
        request_url: str | None = None,
        response_text: str | None = None,
        original_exception: Exception | None = None,
    ):
        self.status_code = status_code
        self.detail = detail
        self.response_text = response_text

        if isinstance(detail, dict) and "detail" in detail:
            error_summary = detail["detail"]
        elif isinstance(detail, str):
            error_summary = detail
        else:
            error_summary = "No detail provided."

        base_message = f"API Error {status_code}: {error_summary}"
        super().__init__(message=base_message, request_url=request_url, original_exception=original_exception)

    def __str__(self):
        message = f"{self.args[0]}"
        if self.request_url:
            message += f"\nRequest URL: {self.request_url}"

        if isinstance(self.detail, dict) and self.detail.get("detail") != self.args[0].split(": ", 1)[-1]:
            full_detail_str = json.dumps(self.detail, indent=2)
            message += f"\nFull API Response Detail:\n{full_detail_str}"
        elif isinstance(self.detail, str) and self.detail != self.args[0].split(": ", 1)[-1]:
            message += f"\nAPI Response Detail: {self.detail}"

        message += f"\n\n{self.extra_info}"
        return message


class BadRequestError(DorsalClientError):
    """400 Bad Request error."""

    def __init__(
        self,
        message: str,
        request_url: str,
        response_text: str | None = None,
        original_exception: Exception | None = None,
    ):
        self.response_text = response_text
        super().__init__(message=message, request_url=request_url, original_exception=original_exception)


class NotFoundError(DorsalClientError):
    """404 Not Found error."""

    def __init__(
        self,
        message: str,
        request_url: str | None = None,
        response_text: str | None = None,
        original_exception: Exception | None = None,
    ):
        self.response_text = response_text
        super().__init__(message=message, request_url=request_url, original_exception=original_exception)


class ForbiddenError(DorsalClientError):
    """403 Forbidden error."""

    def __init__(
        self,
        message: str,
        request_url: str | None = None,
        response_text: str | None = None,
        original_exception: Exception | None = None,
    ):
        self.response_text = response_text
        full_message = f"{message}" if request_url else message
        super().__init__(message=full_message, request_url=request_url, original_exception=original_exception)


class ConflictError(DorsalClientError):
    """409 Conflict error."""

    def __init__(
        self,
        message: str,
        request_url: str | None = None,
        response_text: str | None = None,
        original_exception: Exception | None = None,
    ):
        self.response_text = response_text
        full_message = f"{message}" if request_url else message
        super().__init__(message=full_message, request_url=request_url, original_exception=original_exception)


class RateLimitError(DorsalClientError):
    """429 Too Many Requests error."""

    def __init__(
        self,
        message: str,
        request_url: str,
        retry_after: str | int | None = None,
        response_text: str | None = None,
        original_exception: Exception | None = None,
    ):
        self.retry_after = retry_after
        self.response_text = response_text
        super_message = f"{message} for {request_url}"
        if retry_after:
            super_message += f". Try again after {retry_after} seconds."

        super().__init__(message=super_message, request_url=request_url, original_exception=original_exception)


class ApiDataValidationError(DorsalClientError):
    """API response parsing/validation error."""

    info_url = constants.DOCS_URL_API_ERRORS_VALIDATION
    extra_info_template = (
        f"An error occurred when parsing or validating the API response. \nFor more information, visit: {info_url}"
    )

    def __init__(
        self,
        message: str,
        *,
        request_url: str | None = None,
        validation_errors: list | dict | str | None = None,
        response_text_snippet: str | None = None,
        original_exception: Exception | None = None,
    ):
        super().__init__(message=message, request_url=request_url, original_exception=original_exception)
        self.validation_errors = validation_errors
        self.response_text_snippet = response_text_snippet

    def __str__(self):
        message = f"{self.args[0]}"
        if self.request_url:
            message += f"\nRequest URL: {self.request_url}"
        if self.validation_errors:
            try:
                errors_str = json.dumps(self.validation_errors, indent=2)
            except TypeError:
                errors_str = str(self.validation_errors)
            message += f"\nValidation Issues:\n{errors_str}"
        if self.response_text_snippet:
            message += f"\nProblematic Response Snippet: {self.response_text_snippet}"
        if self.original_exception and not self.validation_errors:
            message += f"\nUnderlying Error: {type(self.original_exception).__name__}: {str(self.original_exception)}"

        message += f"\n\n{self.extra_info_template}"
        return message


# == ModelRunner ==
class ModelRunnerError(DorsalError):
    """Base class for errors specific to the ModelRunner."""

    def __init__(self, message: str, original_exception: Exception | None = None):
        super().__init__(message)
        self.original_exception = original_exception


class ModelRunnerConfigError(ModelRunnerError):
    """Errors related to loading or validating the ModelRunner pipeline configuration."""


class ModelImportError(ModelRunnerError):
    """Errors that occur when trying to import a model or its validator."""

    def __init__(self, callable_path_str: str, original_exception: Exception):
        message = f"Failed to import '{callable_path_str}'"
        super().__init__(message, original_exception)
        self.callable_path_str = callable_path_str


class BaseModelProcessingError(ModelRunnerError):
    """Critical error during the processing of the essential base file model."""


class ModelExecutionError(ModelRunnerError):
    """Error during the execution of an individual annotation model's 'main' method."""

    def __init__(self, model_name: str, file_path: str, original_exception: Exception):
        message = f"Error executing model '{model_name}' for file '{file_path}'"
        super().__init__(message, original_exception)
        self.model_name = model_name
        self.file_path = file_path


class ModelOutputValidationError(ModelRunnerError):
    """Error when a model's output fails Pydantic validation."""

    def __init__(
        self,
        model_name: str,
        validator_name: str,
        errors: list[Any],
        original_exception: Exception | None,
    ):
        message = f"Output from model '{model_name}' failed validation against '{validator_name}'"
        super().__init__(message, original_exception)
        self.model_name = model_name
        self.validator_name = validator_name
        self.validation_errors = errors


class DependencyCheckError(ModelRunnerError):
    """Error occurred within a dependency checker function itself."""

    def __init__(self, checker_path_str: str, original_exception: Exception):
        message = f"Error in dependency checker '{checker_path_str}'"
        super().__init__(message, original_exception)
        self.checker_path_str = checker_path_str


class DependencyNotMetError(ModelRunnerError):
    """A required dependency for a model was not met.
    If this was a non-silent dependency, ModelRunner.run() will propagate this error.
    """

    def __init__(
        self,
        model_name: str,
        dependency_type: str,
        silent: bool,
        details: str | None = None,
    ):
        base_message = f"Dependency '{dependency_type}' not met for model '{model_name}'"
        if details:
            base_message += f": {details}"
        message = f"{base_message} (Silent: {silent})"
        super().__init__(message)
        self.model_name = model_name
        self.dependency_type = dependency_type
        self.silent = silent


class PipelineIntegrityError(ModelRunnerError):
    """Error indicating an unexpected state or structure in the pipeline results during merging."""


class MissingHashError(ModelRunnerError):
    """A required hash (e.g., validation hash) was not found after base model processing."""


# == Hashers ==


class QuickHasherError(DorsalError):
    """Base class for errors specific to QuickHasher operations."""

    def __init__(self, message: str, file_path: str | None = None):
        super().__init__(message)
        self.file_path = file_path
        self.message = message

    def __str__(self) -> str:
        msg = self.message
        if self.file_path:
            msg += f" (File: {self.file_path})"
        return msg


class QuickHashFileInstabilityError(QuickHasherError):
    """Error indicating the file changed or was inconsistent during hashing."""


class QuickHashConfigurationError(QuickHasherError, ValueError):
    """Error related to QuickHasher internal configuration."""


class QuickHashFileSizeError(QuickHasherError, ValueError):
    """Error for file size being outside permitted QuickHasher range when configured to raise."""


# == /file ===


class DuplicateFileError(DorsalError):
    """Duplicate file error."""

    info_url = f"{BASE_URL}/duplicate-files#howto"
    extra_info = (
        f"You cannot index identical files or records at the same time.\n  For more information, visit {info_url}"
    )

    def __init__(self, message, file_paths: list[str] | None = None):
        super().__init__(message)
        self.file_paths = file_paths

    def __str__(self):
        original_message = self.args[0]

        full_message = f"{original_message}"
        if self.file_paths:
            full_message += "\n\nDuplicates:"
            for file_path in self.file_paths:
                full_message += f"\n  {file_path}"
        full_message += f"\n\n{self.extra_info}"
        return full_message


class ReadError(DorsalError):
    """Read error."""


class UnexpectedResponseError(DorsalClientError):
    """The API returned an unexpected response."""


class DatasetExistsError(ConflictError):
    """Dataset exists."""


class PartialIndexingError(DorsalError):
    """An indexing operation completed but includes one or more errors."""

    def __init__(self, message: str, summary: dict, original_error: Exception | None = None):
        super().__init__(message)
        self.summary = summary
        self.original_error = original_error


class BatchIndexingError(DorsalError):
    """Batch indexing operation failure."""

    def __init__(self, message: str, summary: dict, original_error: Exception | None = None):
        super().__init__(message)
        self.summary = summary
        self.original_error = original_error


# == FileAnnotator ==


class FileAnnotatorError(DorsalError):
    """Base exception for all errors originating from the FileAnnotator."""


class AnnotationConfigurationError(FileAnnotatorError):
    """Raised when the configuration for an annotation task is invalid."""


class AnnotationImportError(FileAnnotatorError):
    """Raised when an annotation model or its validator fails to import."""


class AnnotationExecutionError(FileAnnotatorError):
    """Raised when an annotation model fails during execution or its output is invalid."""


class AnnotationValidationError(FileAnnotatorError):
    """Raised when a manually provided annotation fails validation against a schema."""

    def __init__(
        self,
        message: str,
        validation_errors: list | dict | str | None = None,
        original_exception: Exception | None = None,
    ):
        super().__init__(message, original_exception=original_exception)
        self.validation_errors = validation_errors


# File Collections


class SyncConflictError(DorsalError):
    """
    Raised when a sync operation is blocked because the remote collection has changed since the last synchronization.
    """


# Data quality


class DataQualityError(DorsalError):
    """Raised when a data quality rule triggers it."""


# Preprocessing


class PDFProcessingError(DorsalError):
    """Raised when PDF processing fails (corruption, password protection, etc.)."""


class DependencyError(DorsalError):
    """Raised when optional dependencies are missing."""



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from unittest.mock import MagicMock, patch
import tomlkit
from dorsal.api import config as config_api
from dorsal.common.exceptions import DorsalError, DorsalConfigError
from dorsal.common.model import AnnotationModel
from dorsal.file.pipeline_config import PipelineConfig


class MockAnnotationModel(AnnotationModel):
    """A dummy model for registration tests."""

    id = "test/mock"
    version = "0.1.0"

    def main(self):
        return {}


MockAnnotationModel.__module__ = "dorsal.tests.dummy_model"
MockAnnotationModel.__name__ = "MockModel"


@pytest.fixture
def clean_config_environment(tmp_path):
    """
    Redirects config operations to a temporary directory.
    Forces the environment to look like a fresh project with NO config file initially.
    """
    fake_config_path = tmp_path / "dorsal.toml"

    with (
        patch("dorsal.common.config.find_project_config_path", return_value=fake_config_path),
        patch("dorsal.file.pipeline_config.config.find_project_config_path", return_value=fake_config_path),
    ):
        # Ensure we start with a clean slate for the cache
        from dorsal.common.config import load_config

        load_config.cache_clear()

        yield fake_config_path


def test_register_hydrates_defaults_into_empty_config(clean_config_environment):
    """
    Critical Fix Verification:
    Ensures that registering a model into an empty project config
    does NOT wipe out the default pipeline.
    """
    # 1. Register a new model into a non-existent dorsal.toml
    config_api.register_model(annotation_model=MockAnnotationModel, schema_id="test/schema", scope="project")

    # 2. Read the actual file from disk
    with open(clean_config_environment, "r") as f:
        doc = tomlkit.load(f)

    pipeline = doc["model_pipeline"]

    # 3. Assertions
    # We expect the defaults (~5 models) + our new model (1) = ~6 total
    # If the bug existed, len would be 1.
    assert len(pipeline) > 1, "Pipeline defaults were wiped out! Only the new model exists."

    # Verify our model is at the end
    last_step = pipeline[-1]
    assert last_step["annotation_model"] == ["dorsal.tests.dummy_model", "MockModel"]
    assert last_step["schema_id"] == "test/schema"


def test_register_prevents_duplicates(clean_config_environment):
    """Ensures we cannot accidentally add the exact same model twice."""
    # Register once
    config_api.register_model(annotation_model=MockAnnotationModel, schema_id="test/schema", scope="project")

    # Try to register again without overwrite
    with pytest.raises(DorsalConfigError) as exc:
        config_api.register_model(annotation_model=MockAnnotationModel, schema_id="test/schema", scope="project")
    assert "already exists" in str(exc.value)

    # Register with overwrite (should succeed and not increase length)
    config_api.register_model(
        annotation_model=MockAnnotationModel, schema_id="test/schema", overwrite=True, scope="project"
    )

    pipeline = config_api.get_model_pipeline(scope="project")
    # Count occurrences of our mock model
    matches = [s for s in pipeline if s.annotation_model.name == "MockModel"]
    assert len(matches) == 1


def test_base_model_safety(clean_config_environment):
    """Ensures the Base Model (Index 0) cannot be tampered with via API."""
    # Hydrate config first
    config_api.register_model(MockAnnotationModel, "test/schema")

    # 1. Try to Remove Index 0
    with pytest.raises(DorsalError) as exc:
        config_api.remove_model_by_index(0, scope="project")
    assert "Cannot remove the Base Model" in str(exc.value)

    # 2. Try to Deactivate Index 0
    with pytest.raises(DorsalError) as exc:
        config_api.deactivate_model_by_index(0, scope="project")
    assert "Cannot deactivate the Base Model" in str(exc.value)

    # 3. Try to Remove by Name (FileCoreAnnotationModel)
    with pytest.raises(DorsalError) as exc:
        config_api.remove_model_by_name("FileCoreAnnotationModel", scope="project")
    assert "Cannot remove the Base Model" in str(exc.value)


def test_ambiguity_check(clean_config_environment):
    """Ensures generic naming operations fail if names are duplicated."""
    # Manually create a "corrupt" config with duplicates
    # We add MockModel twice manually to bypass the upsert checks
    dup_step = {"annotation_model": ["dorsal.tests", "DuplicateModel"], "schema_id": "test/dup1"}

    # Use internal method to force append twice
    PipelineConfig.upsert_step(dup_step, scope="project")  # Add 1

    # Force add a second one by manually appending to the TOML file
    # (Since upsert_step prevents duplicates normally)
    with open(clean_config_environment, "r") as f:
        doc = tomlkit.load(f)
    doc["model_pipeline"].append(dup_step)
    with open(clean_config_environment, "w") as f:
        tomlkit.dump(doc, f)

    # Now try to act on it by name
    with pytest.raises(ValueError) as exc:
        config_api.activate_model_by_name("DuplicateModel", scope="project")

    assert "Ambiguous model name 'DuplicateModel'" in str(exc.value)
    assert "occurrences at indices" in str(exc.value)


def test_activation_toggles_flag(clean_config_environment):
    """Verifies that deactivate/activate correctly sets the TOML flag."""
    config_api.register_model(MockAnnotationModel, "test/schema")

    # 1. Deactivate
    config_api.deactivate_model_by_name("MockModel", scope="project")

    steps = config_api.get_model_pipeline(scope="project")
    target = next(s for s in steps if s.annotation_model.name == "MockModel")
    assert target.deactivated is True

    # 2. Activate
    config_api.activate_model_by_name("MockModel", scope="project")

    steps = config_api.get_model_pipeline(scope="project")
    target = next(s for s in steps if s.annotation_model.name == "MockModel")
    assert target.deactivated is False


def test_remove_works_with_negative_index(clean_config_environment):
    """Verifies that -1 works as expected."""
    config_api.register_model(MockAnnotationModel, "test/schema")

    # Get current count
    initial_len = len(config_api.get_model_pipeline(scope="project"))

    # Remove last
    config_api.remove_model_by_index(-1, scope="project")

    new_len = len(config_api.get_model_pipeline(scope="project"))
    assert new_len == initial_len - 1



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import re
from typing import Any

from pydantic import BaseModel

from dorsal.common.literals import KB, MB, GB, TB, PB, KiB, MiB, GiB, TiB, PiB

logger = logging.getLogger(__name__)

FILESIZE_UNIT_DP_CONFIG = {"B": 0, "KiB": 0, "MiB": 0, "GiB": 2, "TiB": 2, "PiB": 2}

FILESIZE_ALIAS_MAPPING = {
    "b": 1,
    "byte": 1,
    "bytes": 1,
    "kb": KB,
    "k": KB,
    "kilobyte": KB,
    "kilobytes": KB,
    "mb": MB,
    "m": MB,
    "megabyte": MB,
    "megabytes": MB,
    "gb": GB,
    "g": GB,
    "gigabyte": GB,
    "gigabytes": GB,
    "t": TB,
    "tb": TB,
    "terabyte": TB,
    "terabytes": TB,
    "kib": KiB,
    "kibi": KiB,
    "kibibyte": KiB,
    "kibibytes": KiB,
    "mib": MiB,
    "mibi": MiB,
    "mebibyte": MiB,
    "mebibytes": MiB,
    "gib": GiB,
    "gibi": GiB,
    "gibibyte": GiB,
    "gibibytes": GiB,
    "tib": TiB,
    "tibi": TiB,
    "tibibyte": TiB,
    "tibibytes": TiB,
}


def get_filesize(file_path: str) -> int:
    try:
        return os.path.getsize(file_path)
    except (IOError, PermissionError) as err:
        logger.error("Could not access file at path '%s': %s", file_path, err)
        raise


def human_filesize(filesize: int | float) -> str:
    """Return a file size in a human readable form:

    - To the nearest MiB, GiB, TiB etc.

    In the default config, GiB and higher show 2 decimal places.

    E.g.
        - 22 -> '22 B'
        - 2_233 -> '2 KiB'
        - 223_344 -> '218 KiB'
        - 22_334_455 -> '21 MiB'
        - 2_233_445_566 -> '2.08 GiB'
        - 22_334_455_667_788 -> '20.31 TiB'
        - 2_233_445_566_778_899_001_122 -> '1983698.15 PiB'

    """
    for unit in FILESIZE_UNIT_DP_CONFIG:
        if filesize < 1024.0 or unit == "PiB":
            break
        filesize /= 1024.0
    dp = FILESIZE_UNIT_DP_CONFIG[unit]
    return f"{filesize:.{dp}f} {unit}"


RX_FILESIZE = re.compile(r"^\s*(\d*\.?\d+)\s*([a-zA-Z]+)\s*$")


def parse_filesize(size_str: str) -> int:
    """Parses a human-readable size string (e.g., "10MB", "2.5 GiB") into bytes."""
    size_str = size_str.strip()
    m = RX_FILESIZE.match(size_str)
    if not m:
        try:
            return int(size_str)
        except ValueError:
            logger.debug("Cannot cast %s to integer", size_str)
        logger.warning("Cannot parse file size: Invalid size format: %s", size_str)
        raise ValueError(f"Invalid size format: '{size_str}'")

    value_str, unit = m.groups()
    value = float(value_str)
    unit = unit.lower()

    if unit not in FILESIZE_ALIAS_MAPPING:
        raise ValueError(f"Unknown size unit: '{unit}' in '{size_str}'")

    return int(value * FILESIZE_ALIAS_MAPPING[unit])


def check_record_size(record: dict[str, Any] | BaseModel) -> int:
    """Returns the size of the record in UTF-8 bytes."""
    if isinstance(record, BaseModel):
        data = record.model_dump(mode="json")
    else:
        data = record

    return len(json.dumps(data, separators=(",", ":"), default=str).encode("utf-8"))
